"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
const index_1 = require("../../index");
const internals_fn_1 = require("../Functions/internals.fn");
const lodash_1 = __importDefault(require("lodash"));
const os_1 = __importDefault(require("os"));
const fs_1 = __importDefault(require("fs"));
const fs_extra_1 = __importDefault(require("fs-extra"));
const artisan_fn_1 = __importDefault(require("../Functions/artisan.fn"));
const consoleColors_obj_1 = __importDefault(require("../Objects/consoleColors.obj"));
const Path_1 = __importDefault(require("../Helpers/Path"));
const $ = (0, index_1.getInstance)();
const artisanConfig = $.config.get('artisan', {});
const { logThis, logThisAndExit } = artisan_fn_1.default;
const removeSlashAtEnd = (str) => {
    if (str.slice(-1) === "/") {
        return str.slice(0, str.length - 1);
    }
    return str;
};
const maintenanceFile = $.path.base('.maintenance');
const Commands = {
    up() {
        if (!fs_1.default.existsSync(maintenanceFile)) {
            return $.logAndExit('App is already up');
        }
        fs_1.default.unlinkSync(maintenanceFile);
        $.log('App is now up!');
        $.logAndExit('Reload your server if you have an active getInstance already running.');
    },
    down() {
        if (fs_1.default.existsSync(maintenanceFile)) {
            return $.logAndExit('App is already down.');
        }
        fs_1.default.writeFileSync(maintenanceFile, JSON.stringify({
            date: (new Date()).toUTCString()
        }));
        $.log('App is now in maintenance mood!');
        $.logAndExit('Reload your server if you have an active getInstance already running.');
    },
    install([$plugin]) {
        if ($plugin === "undefined") {
            return logThis("Plugin not specified!");
        }
        const PluginInstaller = require("../Plugins/Installer");
        PluginInstaller($plugin);
    },
    routes([search, query]) {
        if (search && !query) {
            query = search;
            search = 'path';
        }
        if (search === 'ctrl')
            search = 'controller';
        if (search === 'method' && query)
            query = query.toUpperCase();
        let data = [];
        data = $.routerEngine.allProcessedRoutes();
        data.map((e) => {
            e.method = e.method.toUpperCase();
            if (typeof e.controller === "string") {
                const { controller, method } = (0, internals_fn_1.parseControllerString)(e.controller);
                e.controller = `${controller}.${method}`;
            }
            if (typeof e.controller === "function") {
                if (e.controller.name) {
                    e.controller = e.controller.name;
                    if (e.controller.indexOf("getFile(") !== 0) {
                        e.controller += '()';
                    }
                    else {
                        e.controller = e.controller.replace($.path.base(), '');
                    }
                }
                else {
                    e.controller = "anonymous()";
                }
            }
            if (!e.controller)
                e.controller = '';
            if (!e.name)
                e.name = null;
            delete e.url;
        });
        const searchResults = [];
        if (search) {
            for (const route of data) {
                if (!route.hasOwnProperty(search)) {
                    console.log(consoleColors_obj_1.default.fgRed, `Routes table has no column named: (${search}), ACCEPTED: (method | path | controller | name)`);
                    return $.exit();
                }
                if (route[search] && route[search].includes(query)) {
                    searchResults.push(route);
                }
            }
        }
        data = search ? searchResults : data;
        let message = `Total Routes: ${data.length}`;
        if (search)
            message = `Found Routes: ${data.length} WHERE {column: ${search}, query: ${query}}`;
        if (data.length) {
            console.log(consoleColors_obj_1.default.fgCyan);
            console.table(data);
        }
        console.log(consoleColors_obj_1.default.fgYellow, message);
        console.log();
    },
    "make:job"(args) {
        const job = args[0];
        let command = args[1];
        if (typeof job === "undefined") {
            return logThis("Job name not defined!");
        }
        if (typeof command === "undefined") {
            command = job.trim();
        }
        const jobsPath = $.path.backend("jobs");
        artisan_fn_1.default.copyFromFactoryToApp("job", job, jobsPath, { name: job, command }, false);
        return $.exit();
    },
    "make:event"(args) {
        const name = args[0];
        const namespace = args[1];
        const eventsPath = $.path.events();
        artisan_fn_1.default.copyFromFactoryToApp("event", name, eventsPath, { name, namespace }, false);
        return $.exit();
    },
    "make:controller"(args) {
        const controller = args[0];
        if (typeof controller === "undefined") {
            return logThis("Controller name not defined!");
        }
        const controllersPath = removeSlashAtEnd($.path.controllers());
        artisan_fn_1.default.copyFromFactoryToApp("controller", controller, controllersPath);
        return $.exit();
    },
    "make:controller_object"(args) {
        const controller = args[0];
        if (typeof controller === "undefined") {
            return logThis("Controller name not defined!");
        }
        const controllersPath = removeSlashAtEnd($.path.controllers());
        artisan_fn_1.default.copyFromFactoryToApp(["controller", "controller_object"], controller, controllersPath);
        return $.exit();
    },
    "make:controller_services"(args) {
        const controller = args[0];
        if (typeof controller === "undefined") {
            return logThis("Controller name not defined!");
        }
        const controllersPath = removeSlashAtEnd($.path.controllers());
        artisan_fn_1.default.copyFromFactoryToApp(["controller", "controller_with_services"], controller, controllersPath);
        return $.exit();
    },
    "make:controllerService"(args) {
        const service = args[0];
        if (typeof service === "undefined") {
            return logThis("Service name not defined!");
        }
        const controllersPath = removeSlashAtEnd($.path.controllers("services"));
        artisan_fn_1.default.copyFromFactoryToApp(["CService", "controller_service"], service, controllersPath);
        return $.exit();
    },
    "make:middleware"(args) {
        const middleware = args[0];
        if (typeof middleware === "undefined") {
            return logThis("Middleware name not defined!");
        }
        const middlewaresPath = Path_1.default.resolve($.config.get('paths.middlewares'));
        artisan_fn_1.default.copyFromFactoryToApp("middleware", middleware, middlewaresPath);
        return $.exit();
    },
    "make:model"(args) {
        let name = args[0];
        let table = args[1];
        if (typeof name === "undefined") {
            return logThis("Model name not defined!");
        }
        if (typeof table === "undefined") {
            table = lodash_1.default.snakeCase(Path_1.default.path().basename(name));
        }
        if (artisanConfig.singleModelName) {
            name = artisan_fn_1.default.singular(name);
        }
        if (artisanConfig.pluralizeModelTable) {
            table = artisan_fn_1.default.pluralize(table);
        }
        const modelPath = $.path.models();
        artisan_fn_1.default.copyFromFactoryToApp("model", name, modelPath, { name, table });
        $.exit();
    },
    "make:view"(args) {
        const config = $.config.get('template');
        let name = args[0];
        let defaultContent = "";
        if (typeof name === "undefined") {
            return logThis("View name not defined!");
        }
        if (name === "__routes") {
            defaultContent = $.base64.encode($.routerEngine.nameToUrl());
            defaultContent = "<script>" + os_1.default.EOL +
                "window['--routes'] = '" + defaultContent + "';" + os_1.default.EOL +
                "</script>";
        }
        name += "." + config.extension;
        const fullPath = $.path.views(name);
        Path_1.default.makeDirIfNotExist(fullPath, true);
        if (name.substring(0, 2) !== "__" && fs_1.default.existsSync(fullPath)) {
            return logThisAndExit("view {" + consoleColors_obj_1.default.fgYellow + name + consoleColors_obj_1.default.fgCyan + "} already exits!");
        }
        if (!defaultContent.length) {
            const defaultContentFile = $.path.views("_." + config.extension);
            if (fs_1.default.existsSync(defaultContentFile)) {
                defaultContent = fs_1.default.readFileSync(defaultContentFile).toString();
            }
        }
        fs_1.default.writeFileSync(fullPath, defaultContent);
        logThis("View created successfully!");
        logThis("Located @ " + fullPath);
        return $.exit();
    },
    import([plugin, folder, shouldOverwrite]) {
        const overwrite = shouldOverwrite === 'overwrite';
        if (plugin.toLowerCase() === 'xpresser') {
            return this.importFiles(folder, overwrite);
        }
        const allowedImportables = [
            "configs",
            "models",
            "views",
            "events",
            "controllers",
            "middlewares",
        ];
        folder = folder.toLowerCase();
        if (!allowedImportables.includes(folder))
            return $.logErrorAndExit(`Import does not support any importable folder named (${folder})`);
        const config = $.engineData.get(`PluginEngine:namespaces[${plugin}]`);
        if (!config)
            return $.logErrorAndExit(`No plugin namespaced {${plugin}} registered in your project`);
        const importable = config.importable || config.publishable;
        if (!importable)
            return $.logErrorAndExit(`Plugin: {${plugin}} does not have any importables`);
        let importableFactory = importable[folder];
        if (!importableFactory)
            return $.logErrorAndExit(`Plugin: {${plugin}} does not have any importable item named (${folder})`);
        if (importable.hasOwnProperty(folder + '.ts') && $.isTypescript()) {
            importableFactory = importable[folder + '.ts'];
        }
        const from = config.path + '/' + importableFactory;
        if (!fs_1.default.existsSync(from))
            return $.logErrorAndExit(`File/Folder {${importableFactory}} does not exists in plugin (${plugin}) directory.`);
        let to = $.path[folder](lodash_1.default.kebabCase(config.namespace));
        if ($.file.isFile(from)) {
            const ext = Path_1.default.getExtension(from);
            if (ext)
                to += ext;
            Path_1.default.makeDirIfNotExist(to, true);
        }
        else {
            Path_1.default.makeDirIfNotExist(to);
        }
        fs_extra_1.default.copy(from, to, { overwrite })
            .then(() => {
            const base = $.path.base();
            $.logInfo(`From: (${from.replace(base, '')})`);
            $.logInfo(`To: (${to.replace(base, '')})`);
            $.logAndExit('Publish completed!');
        })
            .catch(err => {
            $.logError('An error occurred while publishing the folder.');
            return $.logAndExit(err);
        });
    },
    async importFiles(folder, overwrite) {
        if (folder !== 'types') {
            return $.logErrorAndExit(`Xpresser does not have any importable item named (${folder})`);
        }
        const base = $.path.base();
        const typesDestination = $.path.backend('types');
        const factoryTypesFolder = $.path.engine('Factory/types');
        if (fs_1.default.existsSync(typesDestination)) {
            try {
                for (const type of ['index', 'modules', 'xpresser']) {
                    const from = `${factoryTypesFolder}/${type}.d.ts`;
                    const to = `${typesDestination}/${type}.d.ts`;
                    await fs_extra_1.default.copy(from, to, { overwrite });
                    $.logInfo(`From: (${from.replace(base, '')})`);
                    $.logInfo(`To: (${to.replace(base, '')})`);
                }
                $.logAndExit('Import completed!');
            }
            catch (e) {
                return $.logAndExit(e);
            }
        }
        else {
            $.file.makeDirIfNotExist(typesDestination);
            fs_extra_1.default.copy(factoryTypesFolder, typesDestination, { overwrite })
                .then(() => {
                $.logInfo(`From: (${factoryTypesFolder.replace(base, '')})`);
                $.logInfo(`To: (${typesDestination.replace(base, '')})`);
            })
                .catch(err => {
                $.logError('An error occurred while publishing the folder.');
                return $.logAndExit(err);
            });
        }
    }
};
module.exports = { Commands, Artisan: artisan_fn_1.default };
