"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
const Path_1 = __importDefault(require("./Helpers/Path"));
const has_pkg_1 = __importDefault(require("has-pkg"));
const index_1 = require("../index");
const InXpresserError_1 = __importDefault(require("./Errors/InXpresserError"));
const plugins_fn_1 = require("./Functions/plugins.fn");
const $ = (0, index_1.getInstance)();
const pluginRoutes = [];
const PluginNamespaceToData = {};
class PluginEngine {
    static async loadPlugins(PackageDotJson) {
        const logPlugins = $.config.get("log.plugins", true);
        let plugins;
        const PluginsPath = $.path.jsonConfigs("plugins.json");
        if ($.file.exists(PluginsPath)) {
            try {
                plugins = require(PluginsPath);
            }
            catch (e) {
                $.logError(e);
            }
            if (Array.isArray(plugins) && plugins.length) {
                $.logErrorAndExit('plugins.json should be an object.');
            }
            if (typeof plugins === "object") {
                const pluginKeys = Object.keys(plugins);
                const env = $.config.get('env');
                const pluginPaths = {};
                const pluginData = {};
                const loadedPlugins = {};
                for (const plugin of pluginKeys) {
                    const pluginUseDotJson = plugins[plugin];
                    if (typeof pluginUseDotJson === "boolean" && !pluginUseDotJson)
                        continue;
                    if (typeof pluginUseDotJson === "object") {
                        if (pluginUseDotJson.hasOwnProperty('load') &&
                            pluginUseDotJson.load === false) {
                            continue;
                        }
                        if (pluginUseDotJson.hasOwnProperty('env')) {
                            if (typeof pluginUseDotJson.env === "string" && pluginUseDotJson.env !== env)
                                continue;
                            if (Array.isArray(pluginUseDotJson.env) && !pluginUseDotJson.env.includes(env))
                                continue;
                        }
                    }
                    loadedPlugins[plugin] = pluginUseDotJson;
                }
                const listOfPluginNamespaces = [];
                for (const plugin of Object.keys(loadedPlugins)) {
                    const $pluginPath = pluginPaths[plugin] = Path_1.default.resolve(plugin);
                    try {
                        const $data = pluginData[plugin] = PluginEngine.loadPluginUseData($pluginPath, PackageDotJson);
                        listOfPluginNamespaces.push($data.namespace);
                    }
                    catch (e) {
                        $.logPerLine([
                            { error: plugin },
                            { error: e },
                            { errorAndExit: "" },
                        ], true);
                    }
                }
                $.ifNotConsole(() => {
                    if (logPlugins) {
                        $.logSuccess(`Using plugins: [${listOfPluginNamespaces.join(', ')}]`);
                    }
                    else {
                        const pluginsLength = listOfPluginNamespaces.length;
                        $.logSuccess(`Using (${pluginsLength}) ${pluginsLength === 1 ? 'plugin' : 'plugins'}`);
                    }
                });
                for (const plugin of Object.keys(loadedPlugins)) {
                    if (plugin.length) {
                        const $pluginPath = pluginPaths[plugin];
                        try {
                            const $data = pluginData[plugin];
                            PluginNamespaceToData[$data.namespace] = await PluginEngine.usePlugin(plugin, $pluginPath, $data);
                            $.engineData.set("PluginEngine:namespaces", PluginNamespaceToData);
                        }
                        catch (e) {
                            $.logPerLine([
                                { error: plugin },
                                { error: e },
                                { errorAndExit: "" },
                            ], true);
                        }
                    }
                }
            }
            else {
                $.logWarning('Plugins not loaded! Typeof plugins is expected to be an object.');
            }
        }
        return { routes: pluginRoutes };
    }
    static loadPluginUseData(pluginPath, PackageDotJson) {
        const data = require(pluginPath + "/use.json");
        if (!data.namespace) {
            throw new InXpresserError_1.default(`Cannot read property 'namespace'`);
        }
        if (data.xpresser) {
            let version = data.xpresser;
            const xpresserVersion = PackageDotJson.version;
            const compareWith = version.substring(0, 2);
            version = data.xpresser.substring(2);
            if (compareWith === ">=" && (0, plugins_fn_1.compareVersion)(xpresserVersion, version) === -1) {
                $.logErrorAndExit(`Plugin: [${data.namespace}] requires xpresser version [${compareWith + version}],\nUpgrade xpresser to continue.`);
            }
            else if (compareWith === "<=" &&
                (0, plugins_fn_1.compareVersion)(version, xpresserVersion) === -1) {
                $.logErrorAndExit(`Plugin: [${data.namespace}] requires xpresser version [${compareWith + version}],\nDowngrade xpresser to continue.`);
            }
        }
        return data;
    }
    static async usePlugin(plugin, path, data) {
        const $data = $.objectCollection(data);
        let pluginData;
        pluginData = {
            namespace: $data.get("namespace"),
            plugin,
            path,
            paths: {},
        };
        if ($data.has("paths.routesFile")) {
            let RoutePath = $data.get("paths.routesFile");
            RoutePath = (0, plugins_fn_1.pluginPathExistOrExit)(plugin, path, RoutePath);
            pluginRoutes.push({ plugin, path: RoutePath });
        }
        if ($data.has("paths.controllers")) {
            let controllerPath = $data.get("paths.controllers");
            controllerPath = (0, plugins_fn_1.pluginPathExistOrExit)(plugin, path, controllerPath);
            pluginData.paths.controllers = controllerPath;
        }
        $.ifNotConsole(() => {
            if ($data.has("paths.views")) {
                let viewsPath = $data.get("paths.views");
                viewsPath = (0, plugins_fn_1.pluginPathExistOrExit)(plugin, path, viewsPath);
                pluginData.paths.views = viewsPath;
            }
            if ($data.has("paths.middlewares")) {
                let middlewarePath = $data.get("paths.middlewares");
                middlewarePath = (0, plugins_fn_1.pluginPathExistOrExit)(plugin, path, middlewarePath);
                pluginData.paths.middlewares = middlewarePath;
            }
            if ($data.has("extends")) {
                const extensionData = {};
                if ($data.has("extends.RequestEngine")) {
                    const extenderPath = $data.get("extends.RequestEngine");
                    extensionData["RequestEngine"] = (0, plugins_fn_1.pluginPathExistOrExit)(plugin, path, extenderPath);
                }
                pluginData.extends = extensionData;
            }
        });
        $.ifIsConsole(() => {
            if ($data.has('publishable')) {
                pluginData.publishable = $data.get('publishable');
            }
            if ($data.has('importable')) {
                pluginData.publishable = $data.get('importable');
            }
            if ($data.has("paths.commands")) {
                let commandPath = $data.get("paths.commands");
                commandPath = (0, plugins_fn_1.pluginPathExistOrExit)(plugin, path, commandPath);
                pluginData.paths.commands = commandPath;
                const cliCommandsPath = path + "/cli-commands.json";
                if ($.file.isFile(cliCommandsPath)) {
                    pluginData.commands = {};
                    const cliCommands = require(cliCommandsPath);
                    if (cliCommands && Array.isArray(cliCommands)) {
                        for (const command of cliCommands) {
                            let commandAction = command['action'];
                            if (!commandAction) {
                                commandAction = command.command.split(" ")[0];
                            }
                            pluginData.commands[commandAction] = (0, plugins_fn_1.pluginPathExistOrExit)(plugin, commandPath, command.file);
                        }
                    }
                }
            }
        });
        const useIndex = $data.get('use_index', false);
        if (useIndex) {
            const isCustomFile = typeof useIndex === "string";
            const pluginIndexFile = isCustomFile ? useIndex : 'index.js';
            const indexFilePath = (0, plugins_fn_1.pluginPathExistOrExit)(plugin, path, pluginIndexFile);
            if (indexFilePath) {
                const { run, dependsOn } = require(indexFilePath);
                if (dependsOn && typeof dependsOn === "function") {
                    let pluginDependsOn = await dependsOn(pluginData, $);
                    if (!pluginDependsOn || !Array.isArray(pluginDependsOn))
                        return $.logErrorAndExit(`dependsOn() function for plugin {${pluginData.namespace}} must return an array of packages.`);
                    if (pluginDependsOn.length) {
                        let missingPkgs = 0;
                        pluginDependsOn.forEach(pkg => {
                            if (!(0, has_pkg_1.default)(pkg)) {
                                if (missingPkgs === 0)
                                    $.logError(`Plugin: (${pluginData.namespace}) requires the following dependencies:`);
                                console.log(`- ${pkg}`);
                                missingPkgs++;
                            }
                        });
                        if (missingPkgs)
                            return $.logErrorAndExit(`Install required ${missingPkgs > 1 ? 'dependencies' : 'dependency'} and restart server.`);
                    }
                }
                if (run && typeof run === "function")
                    await run(pluginData, $);
            }
        }
        return pluginData;
    }
}
module.exports = PluginEngine;
