"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
const fs_1 = __importDefault(require("fs"));
const index_1 = require("../index");
const path_1 = __importDefault(require("path"));
const object_collection_1 = __importDefault(require("object-collection"));
const request_fn_1 = __importDefault(require("./Functions/request.fn"));
const ErrorEngine_1 = __importDefault(require("./ErrorEngine"));
const lodash_1 = __importDefault(require("lodash"));
const InXpresserError_1 = __importDefault(require("./Errors/InXpresserError"));
const ejs = require("ejs");
const $ = (0, index_1.getInstance)();
const PluginNameSpaces = $.engineData.get("PluginEngine:namespaces", {});
const useFlash = $.config.get("server.use.flash", false);
const requestEngineConfig = $.config.get('server.requestEngine', {
    dataKey: 'data',
    proceedKey: 'proceed',
    messageKey: 'message'
});
class RequestEngine {
    constructor(req, res, next, route) {
        this.route = {
            name: "",
            method: "",
            controller: "",
        };
        this.res = res;
        this.req = req;
        if (typeof next === "function")
            this.next = next;
        if (route)
            this.route = {
                name: route.name || "",
                method: route.method || "",
                controller: typeof route.controller === "string"
                    ? route.controller : "",
            };
        this.params = req.params || {};
        if (!res.locals)
            res.locals = {};
        this.state = new object_collection_1.default(res.locals);
        this.$body = new object_collection_1.default(req.body || {});
        this.$query = new object_collection_1.default(req.query || {});
    }
    static expressify(fn) {
        return (req, res, next) => {
            return fn(new this(req, res, next));
        };
    }
    $instance() {
        return $;
    }
    $(key) {
        return $[key];
    }
    newError(e) {
        return new ErrorEngine_1.default(this, e);
    }
    next() {
    }
    send(body, status) {
        if (status)
            this.status(status);
        return this.res.send(body);
    }
    json(body, status) {
        if (status)
            this.status(status);
        return this.res.json(body);
    }
    status(code) {
        this.res.status(code);
        return this;
    }
    query(key, $default) {
        return this.$query.get(key, $default);
    }
    body(key, $default) {
        return this.$body.get(key, $default);
    }
    hasParam(param) {
        return this.params.hasOwnProperty(param);
    }
    hasParams(params) {
        if (!Array.isArray(params)) {
            throw new InXpresserError_1.default(`hasParams: Expects argument params to be an array of params.`);
        }
        for (const param of params) {
            if (!this.hasParam(param)) {
                return false;
            }
        }
        return true;
    }
    all(pluck = []) {
        const all = lodash_1.default.extend({}, this.req.query, this.req.body);
        if (pluck.length) {
            return lodash_1.default.pick(all, pluck);
        }
        return all;
    }
    toApi(data = {}, proceed = true, status) {
        const d = { [requestEngineConfig.proceedKey]: proceed };
        if (data.hasOwnProperty(requestEngineConfig.messageKey)) {
            d[requestEngineConfig.messageKey] = data[requestEngineConfig.messageKey];
            delete data[requestEngineConfig.messageKey];
        }
        d.data = data;
        if (status !== undefined)
            this.res.status(status);
        return this.json(d, status);
    }
    toApiFalse(data = {}, status = 200) {
        return this.toApi(data, false, status);
    }
    sayToApi(message, proceed = true, status = 200) {
        return this.toApi({
            [requestEngineConfig.messageKey]: message,
        }, proceed, status);
    }
    sayToApiFalse(message, proceed = false, status = 200) {
        return this.toApi({
            [requestEngineConfig.messageKey]: message,
        }, proceed, status);
    }
    redirect(path = "/") {
        this.res.redirect(path);
        return this.res.end();
    }
    redirectBack() {
        const backURL = this.req.header("Referer") || "/";
        return this.redirect(backURL);
    }
    redirectToRoute(route, keys, query, includeUrl) {
        return this.redirect($.helpers.route(route, keys, query, includeUrl));
    }
    viewData(file) {
        const localsConfig = $.config.get('template.locals');
        const all = localsConfig.all;
        let ctx;
        ctx = lodash_1.default.extend({}, $.helpers, (0, request_fn_1.default)(this));
        ctx.$route = this.route;
        ctx.$currentView = file;
        if (useFlash && this.req.flash) {
            ctx.$flash = this.req.flash();
        }
        ctx.$currentUrl = this.req.url;
        if (all) {
            ctx.$query = this.req.query;
            ctx.$body = this.req.body;
            ctx.$stackedScripts = [];
        }
        else {
            if (localsConfig.stackedScripts)
                ctx.$stackedScripts = [];
            if (localsConfig.query)
                ctx.$query = this.req.query;
            if (localsConfig.body)
                ctx.$body = this.req.body;
        }
        this.res.locals["ctx"] = ctx;
        return ctx;
    }
    view(file, data = {}, fullPath = false, useInternalEjs = false) {
        const defaultRender = (...args) => {
            return this.res.render(...args);
        };
        const Render = typeof this.customRenderer === "function" ? this.customRenderer : defaultRender;
        const $filePath = file;
        if (file.indexOf("::") > 2) {
            if ($.engineData.has("RequestEngine:views." + $filePath)) {
                file = $.engineData.get("RequestEngine:views." + $filePath);
            }
            else {
                const $splitFile = file.split("::");
                const $pluginNamespace = $splitFile[0];
                if (PluginNameSpaces.hasOwnProperty($pluginNamespace)) {
                    const pluginNamespaceData = new object_collection_1.default(PluginNameSpaces[$pluginNamespace]);
                    const pluginViewsPath = pluginNamespaceData.get('paths.views', undefined);
                    if (pluginViewsPath && typeof pluginViewsPath === "string") {
                        file = pluginViewsPath + "/" + $splitFile[1];
                        $.engineData.path("RequestEngine:views").set($filePath, file);
                    }
                }
            }
        }
        const path = file + "." + (useInternalEjs ? "ejs" : $.config.get('template.extension'));
        this.viewData($filePath);
        if (typeof fullPath === "function") {
            return Render(path, data, fullPath);
        }
        if (useInternalEjs === true) {
            data = Object.assign(this.res.locals, data);
            return this.res.send(ejs.render(fs_1.default.readFileSync(path_1.default.resolve(path)).toString(), data, { filename: path }));
        }
        else {
            try {
                return Render(...arguments);
            }
            catch (e) {
                $.logError(e);
            }
        }
    }
    renderView(...args) {
        return this.view(...args);
    }
    render(...args) {
        return this.view(...args);
    }
    renderViewFromEngine(file, data) {
        const view = $.path.engine("backend/views/" + file);
        return this.renderView(view, data, true, true);
    }
    try(fn, log = true) {
        return InXpresserError_1.default.tryOrCatch(fn, e => log ? this.$instance().logError(e) : void 0);
    }
    tryOrCatch(fn, handleError) {
        return InXpresserError_1.default.tryOrCatch(fn, handleError);
    }
    throw(e, log = true) {
        e = InXpresserError_1.default.use(e);
        if (log)
            $.logError(e);
        throw e;
    }
    with(data, value = null) {
        if (this.req.flash) {
            if (typeof data === "string") {
                this.req.flash(data, value);
            }
            else {
                const dataKeys = Object.keys(data);
                for (let i = 0; i < dataKeys.length; i++) {
                    this.req.flash(dataKeys[i], data[dataKeys[i]]);
                }
            }
        }
        return this;
    }
    withOld() {
        if (this.req.flash) {
            const data = this.all();
            const dataKeys = Object.keys(data);
            for (let i = 0; i < dataKeys.length; i++) {
                this.req.flash("old:" + dataKeys[i], data[dataKeys[i]]);
            }
        }
        return this;
    }
    end() {
        return "EndCurrentRequest";
    }
    addToBoot(name, value) {
        this.state.set(`boot.${name}`, value);
        return this;
    }
}
module.exports = RequestEngine;
