"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const path_1 = require("path");
const http_1 = require("http");
const https_1 = require("https");
const index_1 = require("../index");
const chalk_1 = __importDefault(require("chalk"));
const Path_1 = __importDefault(require("./Helpers/Path"));
const OnEventsLoader_1 = require("./Events/OnEventsLoader");
const express_1 = __importDefault(require("express"));
const ExtendedRequestEngine_1 = __importDefault(require("./Plugins/ExtendedRequestEngine"));
const ControllerService_1 = __importDefault(require("./Controllers/ControllerService"));
const $ = (0, index_1.getInstance)();
const isProduction = $.config.get("env") === "production";
const paths = $.config.get("paths");
const useDotJson = $.engineData.get("UseDotJson");
const isUnderMaintenance = $.file.exists($.path.base('.maintenance'));
$.app = (0, express_1.default)();
if (!isProduction) {
    const debugRequest = $.config
        .newInstanceFrom("debug.requests")
        .removeNullOrUndefined()
        .defaults({
        enabled: true,
        colored: true,
        showAll: true,
        show: {
            time: false,
            statusCode: true,
            statusMessage: false
        }
    })
        .all();
    if (debugRequest.enabled) {
        const showAll = debugRequest.showAll === true;
        function getDurationInMilliseconds(start) {
            const NS_PER_SEC = 1e9;
            const NS_TO_MS = 1e6;
            const diff = process.hrtime(start);
            return (diff[0] * NS_PER_SEC + diff[1]) / NS_TO_MS;
        }
        function getStatusColor(message, statusCode) {
            if (statusCode >= 200 && statusCode < 300)
                return chalk_1.default.green(message);
            if (statusCode >= 300 && statusCode < 400)
                return chalk_1.default.dim(message);
            if (statusCode >= 400 && statusCode < 500)
                return chalk_1.default.red(message);
            if (statusCode >= 500 && statusCode < 600)
                return chalk_1.default.blue(message);
            return chalk_1.default.dim(message);
        }
        $.app.use((req, res, next) => {
            const start = process.hrtime();
            res.on("close", () => {
                const duration = getDurationInMilliseconds(start).toLocaleString() + "ms";
                const { method, originalUrl } = req;
                const { statusCode, statusMessage } = res;
                if (debugRequest.ignore && Array.isArray(debugRequest.ignore)) {
                    for (const pattern of debugRequest.ignore) {
                        if (typeof pattern === "number") {
                            if (pattern === res.statusCode)
                                return;
                        }
                        else if (typeof pattern === "string") {
                            if (originalUrl.startsWith(pattern) || originalUrl.includes(pattern))
                                return;
                        }
                        else {
                            if (pattern.test(originalUrl)) {
                                return;
                            }
                        }
                    }
                }
                if (debugRequest.ignoreFn && typeof debugRequest.ignoreFn === "function") {
                    const RequestEngine = $.extendedRequestEngine();
                    if (debugRequest.ignoreFn(new RequestEngine(req, res)))
                        return;
                }
                const time = new Date().toLocaleTimeString();
                let log = "";
                if (showAll || debugRequest.show.time)
                    log += `${time} - `;
                if (showAll || debugRequest.show.statusCode)
                    log += `${statusCode} - `;
                log += `${method} -- ${originalUrl} -- ${duration}`;
                if (showAll || debugRequest.show.statusMessage)
                    log += ` [${statusMessage}]`;
                if (debugRequest.colored) {
                    if (debugRequest.colored === "mute") {
                        console.log(chalk_1.default.dim(log));
                    }
                    else {
                        console.log(getStatusColor(log, statusCode));
                    }
                }
                else {
                    console.log(log);
                }
            });
            next();
        });
    }
}
const forceHttpToHttps = $.config.get("server.ssl.forceHttpToHttps", false);
if (forceHttpToHttps) {
    $.app.use((req, res, next) => {
        const isSecure = req.headers["x-forwarded-proto"] === "https" || req.secure;
        if (isSecure)
            return next();
        let newUrl = `${req.protocol}://${req.hostname}${req.url}`;
        newUrl = newUrl.replace("http://", "https://");
        return res.redirect(newUrl);
    });
}
let poweredBy = $.config.get("server.poweredBy");
if (poweredBy) {
    poweredBy = typeof poweredBy === "string" ? poweredBy : "Xpresser";
    const overrideServerName = $.config.get("response.overrideServerName");
    $.app.use((_req, res, next) => {
        res.set("X-Powered-By", poweredBy);
        if (overrideServerName)
            res.set("Server", poweredBy);
        next();
    });
}
else {
    $.app.disable("x-powered-by");
}
const servePublicFolder = $.config.get("server.servePublicFolder", false);
if (!isUnderMaintenance && servePublicFolder) {
    const responseConfig = $.config.get('response');
    $.app.use(express_1.default.static(paths.public, {
        setHeaders(res, path) {
            if (responseConfig.cacheFiles) {
                if (responseConfig.cacheIfMatch.length) {
                    const match = $.utils.findWordsInString(path, responseConfig.cacheIfMatch);
                    if (match !== null && match.length) {
                        res.set("Cache-Control", "max-age=" + responseConfig.cacheMaxAge);
                    }
                }
                else if (responseConfig.cacheFileExtensions.length) {
                    const files = $.utils.extArrayRegex(responseConfig.cacheFileExtensions);
                    const match = path.match(files);
                    if (match !== null && match.length) {
                        res.set("Cache-Control", "max-age=" + responseConfig.cacheMaxAge);
                    }
                }
            }
        },
    }));
}
const useHelmet = $.config.get("server.use.helmet", isProduction);
if (useHelmet) {
    const helmet = require("helmet");
    const helmetConfig = $.config.get("packages.helmet.config", undefined);
    $.app.use(helmetConfig ? helmet(helmetConfig) : helmet());
    $.logSuccess('Using {Helmet}');
}
const useCors = $.config.get("server.use.cors", false);
if (useCors) {
    const cors = require("cors");
    const corsConfig = $.config.get("packages.cors.config", undefined);
    $.app.use(cors(corsConfig));
}
const useBodyParser = $.config.get("server.use.bodyParser", true);
if (useBodyParser) {
    const bodyParser = require("body-parser");
    const bodyParserJsonConfig = $.config.get("packages.body-parser.json");
    const bodyParserUrlEncodedConfig = $.config.get("packages.body-parser.urlencoded", { extended: true });
    $.app.use(bodyParser.json(bodyParserJsonConfig));
    $.app.use(bodyParser.urlencoded(bodyParserUrlEncodedConfig));
    $.app.use((err, req, res, next) => {
        if (err && err["type"] && err["type"] === "entity.parse.failed") {
            return next();
        }
        return next(err);
    });
}
const useFlash = $.config.get('server.use.flash', false);
if (useFlash) {
    const flash = require("express-flash");
    $.app.use(flash());
}
const template = $.config.get('template');
if (typeof template.engine === "function") {
    $.app.engine(template.extension, template.engine);
    $.app.set("view engine", template.extension);
}
else {
    if (typeof template.use === "string") {
        $.app.use($.use.package(template.use));
    }
    else if (typeof template.use === "function") {
        $.app.use(template.use);
    }
    else {
        $.app.set("view engine", template.engine);
    }
}
$.app.set("views", $.path.views());
if (isUnderMaintenance) {
    $.logError(`${$.config.get("name")}, is under Maintenance!`);
    let maintenanceMiddleware = $.config.get('server.maintenanceMiddleware');
    maintenanceMiddleware = $.path.middlewares(maintenanceMiddleware);
    const maintenanceMiddlewareExists = $.file.exists(maintenanceMiddleware);
    if (maintenanceMiddlewareExists)
        maintenanceMiddleware = require(maintenanceMiddleware);
    $.app.use(ExtendedRequestEngine_1.default.expressify(http => {
        if (maintenanceMiddlewareExists)
            return maintenanceMiddleware(http);
        return http.newError().view({
            error: {
                title: 'Maintenance Mood!',
                log: 'We will be right back shortly!',
            },
        }, 200);
    }));
}
const convertBodyEmptyStringToNull = $.config.get('server.convertBodyEmptyStringToNull', true);
if (convertBodyEmptyStringToNull) {
    $.app.use((req, _res, next) => {
        if (req.body && Object.keys(req.body).length) {
            for (const [key, value] of Object.entries(req.body)) {
                if (typeof value === "string" && value.trim() === "") {
                    req.body[key] = null;
                }
            }
        }
        return next();
    });
}
const afterExpressInit = (next) => {
    if (!$.options.isTinker) {
        $.controller = require("./Classes/ControllerClass");
        $.handler = (controller) => new ControllerService_1.default(controller);
    }
    if (useDotJson.has("globalMiddlewares")) {
        const projectFileExtension = $.config.get('project.fileExtension');
        const $middlewares = useDotJson.get("globalMiddlewares");
        for (let i = 0; i < $middlewares.length; i++) {
            let $middleware = $middlewares[i];
            if ($middleware.slice(-3) !== projectFileExtension) {
                $middleware += projectFileExtension;
            }
            $middleware = Path_1.default.resolve($middleware);
            try {
                $.app.use(ExtendedRequestEngine_1.default.expressify(require($middleware)));
            }
            catch (e) {
                $.logPerLine([
                    { error: "Error in use.json" },
                    { error: e },
                    { errorAndExit: "" },
                ]);
            }
        }
    }
    require("./Routes/Loader");
    $.routerEngine.processRoutes($.router.routes);
    return next();
};
const startHttpServer = (onSuccess, onError) => {
    $.app.use(ExtendedRequestEngine_1.default.expressify(http => {
        http.res.status(404);
        if (http.req.xhr) {
            return http.send({ error: "Not found" });
        }
        else {
            return http.newError().pageNotFound();
        }
    }));
    $.http = (0, http_1.createServer)($.app);
    const port = $.config.get("server.port", 80);
    $.http.on("error", (err) => {
        if (err["errno"] === "EADDRINUSE") {
            return $.logErrorAndExit(`Port ${err["port"]} is already in use.`);
        }
        if (typeof onError === "function") {
            onError(err);
        }
    });
    (0, OnEventsLoader_1.runBootEvent)("http", () => {
        $.http.listen(port, () => {
            const serverDomainAndPort = $.config.get("log.serverDomainAndPort");
            const domain = $.config.get('server.domain');
            const baseUrl = $.helpers.url().trim();
            const lanIp = $.engineData.get("lanIp");
            const ServerStarted = new Date();
            const getServerUptime = () => global.moment(ServerStarted).fromNow();
            if (serverDomainAndPort || (baseUrl === '' || baseUrl === '/')) {
                $.log(`Domain: ${domain} | Port: ${port} | BaseUrl: ${baseUrl}`);
            }
            else {
                $.log(`Url: ${baseUrl}`);
            }
            if (!isProduction && lanIp)
                $.log(`Network: http://${lanIp}:${port}/`);
            if (isProduction)
                $.log(`Server started - ${ServerStarted.toString()}`);
            $.engineData.set({
                ServerStarted,
                getServerUptime,
                lanIp
            });
            const hasSslEnabled = $.config.get("server.ssl.enabled", false);
            if (hasSslEnabled) {
                startHttpsServer();
            }
            else {
                (0, OnEventsLoader_1.runBootEvent)('serverBooted');
            }
            if (typeof onSuccess === "function") {
                onSuccess();
            }
        });
    });
};
const startHttpsServer = () => {
    const httpsPort = $.config.get("server.ssl.port", 443);
    if (!$.config.has("server.ssl.files")) {
        $.logErrorAndExit("Ssl enabled but has no {server.ssl.files} config found.");
    }
    const files = $.config.get("server.ssl.files");
    if (typeof files.key !== "string" || typeof files.cert !== "string") {
        $.logErrorAndExit("Config {server.ssl.files} not configured properly!");
    }
    if (!files.key.length || !files.cert.length) {
        $.logErrorAndExit("Config {server.ssl.files} not configured properly!");
    }
    files.key = (0, path_1.resolve)(files.key);
    files.cert = (0, path_1.resolve)(files.cert);
    if (!$.file.exists(files.key)) {
        $.logErrorAndExit("Key file {" + files.key + "} not found!");
    }
    if (!$.file.exists(files.cert)) {
        $.logErrorAndExit("Cert file {" + files.key + "} not found!");
    }
    files.key = $.file.read(files.key).toString();
    files.cert = $.file.read(files.cert).toString();
    $.https = (0, https_1.createServer)(files, $.app);
    $.https.on("error", $.logError);
    (0, OnEventsLoader_1.runBootEvent)("https", () => {
        $.https.listen(httpsPort, () => {
            $.logSuccess("Ssl Enabled.");
            (0, OnEventsLoader_1.runBootEvent)('serverBooted');
        });
    });
};
(0, OnEventsLoader_1.runBootEvent)("expressInit", () => afterExpressInit(() => (0, OnEventsLoader_1.runBootEvent)("bootServer", startHttpServer)));
