"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
const path_1 = __importDefault(require("path"));
const fs_1 = __importDefault(require("fs"));
class XpresserRepl {
    constructor(config) {
        this.started = false;
        this.data = {
            commandPrefix: 'xpresser>',
            configProvider() {
                return {};
            },
        };
        this.context = {};
        if (config)
            this.setConfigProvider(config);
    }
    hasXpresserProvider() {
        return typeof this.data.xpresserProvider === "function";
    }
    hasBeforeStartFn() {
        return typeof this.data.beforeStart === "function";
    }
    setConfigProvider(configProvider) {
        const fs = require('fs');
        const path = require('path');
        if (typeof configProvider === "string") {
            const configPath = path.resolve(configProvider);
            if (!fs.existsSync(configPath)) {
                throw Error(`Config path not found! {${configPath}}`);
            }
            this.data.configProvider = () => require(configPath);
        }
        else if (typeof configProvider === "function") {
            this.data.configProvider = configProvider;
        }
        return this;
    }
    setXpresserProvider(xpresserProvider) {
        this.data.xpresserProvider = xpresserProvider;
        return this;
    }
    setCommandPrefix(commandPrefix) {
        this.data.commandPrefix = commandPrefix;
        return this;
    }
    beforeStart(run) {
        this.data.beforeStart = run;
        return this;
    }
    addContext(key, value) {
        if (this.started) {
            throw Error(`addContext(): cannot be used after repl server has started, use replServer.context instead`);
        }
        if (typeof key === "string") {
            this.context[key] = value;
        }
        else if (typeof key === "object") {
            Object.assign(this.context, key);
        }
        return this;
    }
    async buildInstance() {
        const { init } = require('xpresser');
        const config = await this.data.configProvider();
        return init(config, {
            requireOnly: true,
            isConsole: true
        });
    }
    async start(onReplStart) {
        const repl = require('repl');
        const chalk = require('chalk');
        const _ = require('lodash');
        let xpr;
        if (this.hasXpresserProvider()) {
            xpr = await this.data.xpresserProvider();
            if (!xpr.options.requireOnly)
                xpr.options.requireOnly = true;
            if (!xpr.options.isConsole)
                xpr.options.isConsole = true;
        }
        else {
            xpr = await this.buildInstance();
        }
        if (!xpr) {
            throw Error('Xpresser instance is undefined!');
        }
        xpr.on.boot(async (next) => {
            this.context.$ = xpr;
            console.log(chalk.gray('>>>>>>>>>> >>>>>>>>>> >>>>>>>>>> >>>>>>>>>> >>>>>>>>>>'));
            console.log(chalk.white(`Xpresser Repl Session.`));
            console.log();
            console.log(`Env: ${chalk.yellow(xpr.config.get('env'))}`);
            console.log(`Name: ${chalk.yellow(xpr.config.get('name'))}`);
            console.log();
            console.log(chalk.white(`Use ${chalk.whiteBright('.end')} to end repl session.`));
            console.log(chalk.gray('<<<<<<<<<< <<<<<<<<<< <<<<<<<<<< <<<<<<<<<< <<<<<<<<<<'));
            if (this.hasBeforeStartFn()) {
                await this.data.beforeStart(xpr);
            }
            return next();
        });
        xpr.on.boot(async () => {
            this.server = repl.start({
                prompt: chalk.cyanBright(`${this.data.commandPrefix.trim()} `),
                useColors: true,
                terminal: true,
            });
            const replHistory = this.data.historyFile || xpr.path.storage('framework/.repl_history');
            if (!fs_1.default.existsSync(replHistory)) {
                xpr.file.makeDirIfNotExist(replHistory, true);
            }
            this.server.setupHistory(replHistory, (err) => {
                if (err)
                    throw err;
            });
            this.server.defineCommand('end', {
                help: 'End repl session.',
                action() {
                    xpr.log("Goodbye! See you later...");
                    xpr.exit();
                }
            });
            this.server.defineCommand('clearHistory', {
                help: 'Clear repl history.',
                action() {
                    try {
                        fs_1.default.unlinkSync(replHistory);
                        xpr.log("History cleared, restart repl!");
                        xpr.exit();
                    }
                    catch (e) {
                        xpr.logError(e);
                    }
                }
            });
            this.server.defineCommand('ls', {
                help: 'List context in current session.',
                action: (context) => {
                    if (context) {
                        if (!_.has(this.context, context)) {
                            xpr.logError(`No context with name: ${context}`);
                            return this.server.displayPrompt();
                        }
                        const contextValue = _.get(this.context, context);
                        if (typeof contextValue === "object") {
                            mapAndLogContext(chalk, xpr, contextValue);
                            return this.server.displayPrompt();
                        }
                        console.log(`${chalk.yellowBright(context)}:`, typeof (contextValue));
                        return this.server.displayPrompt();
                    }
                    mapAndLogContext(chalk, xpr, this.context);
                    this.server.displayPrompt();
                }
            });
            if (typeof onReplStart === "function") {
                await onReplStart(xpr);
            }
            Object.assign(this.server.context, this.context);
            this.started = true;
        });
        xpr.boot();
        return this;
    }
    addContextFromFiles(files, as, interceptor) {
        const contentNames = Object.keys(files);
        const context = {};
        for (const contextName of contentNames) {
            try {
                const file = path_1.default.resolve(files[contextName]);
                context[contextName] = interceptor ? interceptor(require(file)) : require(file);
            }
            catch (e) {
                throw e;
            }
        }
        return as ? this.addContext(as, context) : this.addContext(context);
    }
    addContextFromFolder(folder, as, extensions, interceptor) {
        const { getAllFiles } = require("./Functions/inbuilt.fn");
        const { camelCase } = require('lodash');
        folder = path_1.default.resolve(folder);
        let allowedExtensions = ['js', 'ts'];
        if (extensions)
            allowedExtensions = allowedExtensions.concat(extensions);
        const files = getAllFiles(folder);
        const context = {};
        for (const file of files) {
            const extension = file.split('.').pop();
            if (allowedExtensions.includes(extension)) {
                let shortFilePath = file.replace(folder, '');
                shortFilePath = shortFilePath.substring(0, shortFilePath.length - `.${extension}`.length);
                const contextName = capitalize(camelCase(shortFilePath));
                try {
                    context[contextName] = interceptor ? interceptor(require(file)) : require(file);
                }
                catch (e) {
                    throw e;
                }
            }
        }
        return as ? this.addContext(as, context) : this.addContext(context);
    }
}
const capitalize = (s) => {
    return s.charAt(0).toUpperCase() + s.slice(1);
};
const getTypeOfContext = (context) => {
    if (context) {
        if (typeof context === 'object' && context.constructor && context.constructor.name) {
            return context.constructor.name;
        }
        else if (typeof context === 'function' && context.name) {
            return context.name;
        }
    }
    return typeof context;
};
const mapAndLogContext = (chalk, xpr, entries) => {
    Object.entries(entries).forEach(([entry, value]) => {
        if (value === xpr) {
            console.log(`${chalk.yellowBright(entry)}:`, 'XpresserInstance ($)');
        }
        else {
            console.log(`${chalk.yellowBright(entry)}:`, getTypeOfContext(value));
        }
    });
};
module.exports = XpresserRepl;
